/*
 * File...........: linux/fs/partitions/ibm.c      
 * Author(s)......: Holger Smolinski <Holger.Smolinski@de.ibm.com>
 *                  Volker Sameske <sameske@de.ibm.com>
 * Bugreports.to..: <Linux390@de.ibm.com>
 * (C) IBM Corporation, IBM Deutschland Entwicklung GmbH, 1999,2000

 * History of changes (starts July 2000)
 * 07/10/00 Fixed detection of CMS formatted disks     
 * 02/13/00 VTOC partition support added
 * 12/27/01 fixed PL030593 (CMS reserved minidisk not detected on 64 bit)
 */

#include <linux/config.h>
#include <linux/fs.h>
#include <linux/genhd.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/blk.h>
#include <linux/slab.h>
#include <linux/hdreg.h>
#include <linux/ioctl.h>
#include <linux/version.h>
#include <asm/ebcdic.h>
#include <asm/uaccess.h>
#include <asm/dasd.h>

#include "ibm.h"
#include "check.h"
#include <asm/vtoc.h>

extern void md_autodetect_dev(kdev_t dev);

static int
get_drive_geometry(struct block_device *bdev, struct hd_geometry *geo) 
{
	int rc;

	rc = blkdev_get(bdev, 0, 1, BDEV_FILE);
	if (rc == 0) {
		rc = ioctl_by_bdev(bdev, HDIO_GETGEO, (unsigned long) geo);
		blkdev_put(bdev, BDEV_FILE);
	}
	return rc;
}

static int
get_drive_info(struct block_device *bdev, dasd_information_t *info) 
{
	int rc;

	rc = blkdev_get(bdev, 0, 1, BDEV_FILE);
	if (rc == 0) {
		rc = ioctl_by_bdev(bdev, BIODASDINFO, (unsigned long) info);
		blkdev_put(bdev, BDEV_FILE);
	}
	return rc;
}

/*
 * compute the block number from a 
 * cyl-cyl-head-head structure
 */
static inline int
cchh2blk (cchh_t *ptr, struct hd_geometry *geo) {
        return ptr->cc * geo->heads * geo->sectors +
	       ptr->hh * geo->sectors;
}


/*
 * compute the block number from a 
 * cyl-cyl-head-head-block structure
 */
static inline int
cchhb2blk (cchhb_t *ptr, struct hd_geometry *geo) {
        return ptr->cc * geo->heads * geo->sectors +
		ptr->hh * geo->sectors +
		ptr->b;
}

int 
ibm_partition(struct gendisk *hd, kdev_t dev, unsigned long first_sector, int
first_part_minor)
{
	struct block_device *bdev = bdget(dev);
	int blocksize, offset, size;
	dasd_information_t info;
	struct hd_geometry geo;
	struct buffer_head *bh;
	char type[5] = {0,};
	char name[7] = {0,};
	volume_label_t vlabel;

	if ( first_sector != 0 )
		BUG();

	if (get_drive_info(bdev, &info) != 0 ||
	    get_drive_geometry(bdev, &geo) != 0 ||
	    hardsect_size[MAJOR(dev)][MINOR(dev)] <= 0)
		return 0;
	blocksize = hardsect_size[MAJOR(dev)][MINOR(dev)];

	/*
	 * Get volume label, extract name and type.
	 */
	set_blocksize(dev, blocksize);  /* OUCH !! */
	if ((bh = bread(dev, info.label_block, blocksize)) == NULL)
		return 0;
	strncpy (type, bh->b_data, 4);
	if ((!info.FBA_layout) && (!strcmp(info.type, "ECKD")))
		strncpy(name, bh->b_data + 8, 6);
	else
		strncpy(name, bh->b_data + 4, 6);
	memcpy (&vlabel, bh->b_data, sizeof(volume_label_t));
	bforget(bh);

	EBCASC(type, 4);
	EBCASC(name, 6);

	/*
	 * Three different types: CMS1, VOL1 and LNX1/unlabeled
	 */
	if (strncmp(type, "CMS1", 4) == 0) {
		/*
		 * VM style CMS1 labeled disk
		 */
		int *label = (int *) bh->b_data;

		if (label[13] != 0) {
			printk("CMS1/%8s(MDSK):", name);
			/* disk is reserved minidisk */
			blocksize = label[3];
			offset = label[13];
			size = (label[7] - 1)*(blocksize >> 9);
		} else {
			printk("CMS1/%8s:", name);
			offset = (info.label_block + 1);
			size = hd->sizes[MINOR(dev)] << 1;
		}
		add_gd_partition( hd, MINOR(dev), 0, size);
		add_gd_partition( hd, MINOR(dev)+1, offset*(blocksize >> 9),
				  size-offset*(blocksize >> 9));
	} else if (strncmp(type, "VOL1", 4) == 0) {
		/*
		 * New style VOL1 labeled disk
		 */
		unsigned int blk;
		int counter;

		printk("VOL1/%8s:", name);
		add_gd_partition(hd, MINOR(dev), 0, hd->sizes[MINOR(dev)]<<1);
		
		/* get block number and read then go through format1 labels */
		blk = cchhb2blk(&vlabel.vtoc, &geo) + 1;
		counter = 1;
		while ((bh = bread(dev, blk, blocksize)) != NULL) {
			format1_label_t f1;

			memcpy(&f1, bh->b_data, sizeof(format1_label_t));
			bforget(bh);

			/* skip FMT4 / FMT5 / FMT7 labels */
			if (f1.DS1FMTID == _ascebc['4']
			    || f1.DS1FMTID == _ascebc['5']
			    || f1.DS1FMTID == _ascebc['7']) {
			        blk++;
				continue;
			}

			/* only FMT1 valid at this point */
			if (f1.DS1FMTID != _ascebc['1'])
				break;

			/* OK, we got valid partition data */
		        offset = cchh2blk(&f1.DS1EXT1.llimit, &geo);
			size  = cchh2blk(&f1.DS1EXT1.ulimit, &geo) - 
				offset + geo.sectors;
			if (counter >= hd->max_p)
				break;
			add_gd_partition(hd, MINOR(dev) + counter, 
					 offset * (blocksize >> 9),
					 size * (blocksize >> 9));


#if CONFIG_BLK_DEV_MD
			{
			  char ptype[45];
			  char *ch;
			  strncpy (ptype, f1.DS1DSNAM, 44);
			  ptype[44] = 0;
			  EBCASC(ptype, 44);
			  ch = strstr(ptype, "PART");
			  if (ch != NULL) {
			    if(strncmp(ch + 9, "RAID  ", 6) == 0) {
			      md_autodetect_dev(MKDEV(hd->major,
						      MINOR(dev) + counter));
			    }
			  }
			}
#endif

			counter++;
			blk++;
		}
	} else {
		/*
		 * Old style LNX1 or unlabeled disk
		 */
		if (strncmp(type, "LNX1", 4) == 0)
			printk ("LNX1/%8s:", name);
		else
			printk("(nonl)/%8s:", name);
		offset = (info.label_block + 1);
		size = hd->sizes[MINOR(dev)] << 1;
		add_gd_partition( hd, MINOR(dev), 0, size);
		add_gd_partition( hd, MINOR(dev)+1, offset*(blocksize >> 9),
				  size-offset*(blocksize >> 9));
	}

	printk("\n");
	return 1;
}
